"""Record runtime learning packet applications."""

from __future__ import annotations

from kb_mcp.events.store import EventStore
from kb_mcp.note import generate_ulid


def record_learning_application(
    *,
    packet_id: str,
    tool_name: str,
    tool_call_id: str,
    source_tool: str,
    source_client: str,
    session_id: str | None,
    save_request_id: str | None,
    saved_note_id: str | None,
    saved_note_path: str | None,
    store: EventStore | None = None,
) -> str:
    application_id = generate_ulid()
    (store or EventStore()).record_learning_application(
        application_id=application_id,
        packet_id=packet_id,
        tool_name=tool_name,
        tool_call_id=tool_call_id,
        source_tool=source_tool,
        source_client=source_client,
        session_id=session_id,
        save_request_id=save_request_id,
        saved_note_id=saved_note_id,
        saved_note_path=saved_note_path,
    )
    return application_id



"""Build runtime learning packets from resolved assets."""

from __future__ import annotations

from kb_mcp.events.store import EventStore
from kb_mcp.learning.models import LearningAssetView, ResolverInput
from kb_mcp.learning.revocation import packet_expires_at
from kb_mcp.note import generate_ulid


def build_learning_packet(
    request: ResolverInput,
    *,
    tool_name: str,
    assets: list[LearningAssetView],
    store: EventStore | None = None,
) -> dict[str, object] | None:
    if not assets:
        return None
    packet_id = generate_ulid()
    active_store = store or EventStore()
    asset_keys = [asset.asset_key for asset in assets]
    active_store.create_learning_packet(
        packet_id=packet_id,
        source_tool=request.source_tool,
        source_client=request.source_client,
        tool_name=tool_name,
        session_id=request.session_id,
        project=request.project,
        repo=request.repo,
        cwd=request.cwd,
        asset_keys=asset_keys,
        expires_at=packet_expires_at(),
    )
    return {
        "packet_id": packet_id,
        "asset_keys": asset_keys,
        "assets": assets,
    }



"""Request-context wrapper for kb-owned MCP tool events."""

from __future__ import annotations

import os
import inspect
from typing import Any, Callable, TypeVar
from uuid import uuid4

from mcp.server.fastmcp import Context

from kb_mcp.events.request_context import REQUEST_CONTEXT
from kb_mcp.events.normalize import normalize_event
from kb_mcp.events.store import EventStore
from kb_mcp.events.worker import run_once
from kb_mcp.learning.application_trace import record_learning_application
from kb_mcp.learning.models import ResolverInput
from kb_mcp.learning.packet_builder import build_learning_packet
from kb_mcp.learning.resolver import resolve_learning_assets

F = TypeVar("F", bound=Callable[..., Any])

def _build_request_context(tool_name: str, ctx: Context | None) -> dict[str, Any]:
    request_id = None
    meta = None
    if ctx is not None:
        request_id = str(ctx.request_id)
        meta = ctx.request_context.meta
    correlation_id = os.environ.get("KB_SESSION_CORRELATION_ID") or None
    vendor_session_id = os.environ.get("KB_VENDOR_SESSION_ID") or None
    tool_call_id = request_id or str(uuid4())
    return {
        "tool_name": tool_name,
        "request_id": request_id,
        "meta": meta,
        "correlation_id": correlation_id,
        "vendor_session_id": vendor_session_id,
        "tool_call_id": tool_call_id,
        "save_request_id": str(uuid4()) if tool_name in {"adr", "gap", "knowledge", "session", "draft"} else None,
    }


def _prepare_learning_packet(
    *,
    source_tool: str,
    source_client: str,
    tool_name: str,
    call_kwargs: dict[str, Any],
) -> dict[str, Any]:
    current = dict(REQUEST_CONTEXT.get() or {})
    request = ResolverInput(
        source_tool=source_tool,
        source_client=source_client,
        session_id=current.get("vendor_session_id"),
        project=call_kwargs.get("project"),
        cwd=call_kwargs.get("cwd"),
        repo=call_kwargs.get("repo"),
    )
    assets = resolve_learning_assets(request)
    packet = build_learning_packet(request, tool_name=tool_name, assets=assets)
    if packet is None:
        return current
    current["packet_id"] = packet["packet_id"]
    current["applied_asset_keys"] = packet["asset_keys"]
    return current


def emit_tool_event(
    *,
    source_tool: str,
    source_client: str,
    event_name: str,
    tool_name: str,
    ctx: Context | None = None,
    payload: dict[str, Any] | None = None,
) -> None:
    """Emit an authoritative tool event."""
    request_context = REQUEST_CONTEXT.get() or _build_request_context(tool_name, ctx)
    body = {
        "tool_name": tool_name,
        "tool_call_id": request_context["tool_call_id"],
        "session_id": request_context["vendor_session_id"],
    }
    if request_context.get("packet_id"):
        body["packet_id"] = request_context["packet_id"]
        body["applied_asset_keys"] = list(request_context.get("applied_asset_keys") or [])
    if payload:
        body.update(payload)
    envelope = normalize_event(
        tool=source_tool,
        client=source_client,
        layer="server_middleware",
        event=event_name,
        payload=body,
    )
    EventStore().append(envelope)
    run_once()


def with_tool_events(source_tool: str, source_client: str, tool_name: str, fn: F) -> F:
    """Wrap a tool handler with before/after/error event emission."""

    if inspect.iscoroutinefunction(fn):
        async def wrapped_async(*args: Any, **kwargs: Any) -> Any:
            call_kwargs = dict(kwargs)
            ctx = call_kwargs.pop("ctx", None)
            request_context = _build_request_context(tool_name, ctx)
            token = REQUEST_CONTEXT.set(request_context)
            REQUEST_CONTEXT.set(
                _prepare_learning_packet(
                    source_tool=source_tool,
                    source_client=source_client,
                    tool_name=tool_name,
                    call_kwargs=call_kwargs,
                )
            )
            emit_tool_event(
                source_tool=source_tool,
                source_client=source_client,
                event_name="tool_started",
                tool_name=tool_name,
                ctx=ctx,
            )
            try:
                result = await fn(*args, **call_kwargs)
            except Exception as exc:
                emit_tool_event(
                    source_tool=source_tool,
                    source_client=source_client,
                    event_name="tool_failed",
                    tool_name=tool_name,
                    ctx=ctx,
                    payload={"message": str(exc)},
                )
                REQUEST_CONTEXT.reset(token)
                raise
            emit_tool_event(
                source_tool=source_tool,
                source_client=source_client,
                event_name="tool_succeeded",
                tool_name=tool_name,
                ctx=ctx,
                payload=_success_payload(),
            )
            _record_learning_application_from_context(
                source_tool=source_tool,
                source_client=source_client,
                tool_name=tool_name,
            )
            REQUEST_CONTEXT.reset(token)
            return result

        return wrapped_async  # type: ignore[return-value]

    def wrapped(*args: Any, **kwargs: Any) -> Any:
        call_kwargs = dict(kwargs)
        ctx = call_kwargs.pop("ctx", None)
        request_context = _build_request_context(tool_name, ctx)
        token = REQUEST_CONTEXT.set(request_context)
        REQUEST_CONTEXT.set(
            _prepare_learning_packet(
                source_tool=source_tool,
                source_client=source_client,
                tool_name=tool_name,
                call_kwargs=call_kwargs,
            )
        )
        emit_tool_event(
            source_tool=source_tool,
            source_client=source_client,
            event_name="tool_started",
            tool_name=tool_name,
            ctx=ctx,
        )
        try:
            result = fn(*args, **call_kwargs)
        except Exception as exc:
            emit_tool_event(
                source_tool=source_tool,
                source_client=source_client,
                event_name="tool_failed",
                tool_name=tool_name,
                ctx=ctx,
                payload={"message": str(exc)},
            )
            REQUEST_CONTEXT.reset(token)
            raise
        emit_tool_event(
            source_tool=source_tool,
            source_client=source_client,
            event_name="tool_succeeded",
            tool_name=tool_name,
            ctx=ctx,
            payload=_success_payload(),
        )
        _record_learning_application_from_context(
            source_tool=source_tool,
            source_client=source_client,
            tool_name=tool_name,
        )
        REQUEST_CONTEXT.reset(token)
        return result

    return wrapped  # type: ignore[return-value]


def _success_payload() -> dict[str, Any]:
    current = REQUEST_CONTEXT.get() or {}
    payload: dict[str, Any] = {}
    for key in ["save_request_id", "saved_note_id", "saved_note_path", "saved_note_type"]:
        value = current.get(key)
        if value:
            payload[key] = value
    return payload


def _record_learning_application_from_context(
    *,
    source_tool: str,
    source_client: str,
    tool_name: str,
) -> None:
    current = REQUEST_CONTEXT.get() or {}
    packet_id = current.get("packet_id")
    tool_call_id = current.get("tool_call_id")
    if not packet_id or not tool_call_id:
        return
    record_learning_application(
        packet_id=str(packet_id),
        tool_name=tool_name,
        tool_call_id=str(tool_call_id),
        source_tool=source_tool,
        source_client=source_client,
        session_id=current.get("vendor_session_id"),
        save_request_id=current.get("save_request_id"),
        saved_note_id=current.get("saved_note_id"),
        saved_note_path=current.get("saved_note_path"),
    )



"""Typed models for learning asset resolution."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass(slots=True)
class ResolverInput:
    source_tool: str
    source_client: str
    session_id: str | None = None
    project: str | None = None
    cwd: str | None = None
    repo: str | None = None


@dataclass(slots=True)
class LearningAssetView:
    asset_key: str
    memory_class: str
    update_target: str
    scope: str
    force: str
    confidence: str
    lifecycle: str
    learning_state_visibility: str
    candidate_key: str | None
    review_id: str | None
    materialization_key: str | None
    note_id: str | None
    note_path: str | None
    updated_at: str
    provenance: dict[str, object]
    traceability: dict[str, object]
    revocation_path: dict[str, object]



from __future__ import annotations

import os
import tempfile
import unittest
from pathlib import Path

import yaml

from kb_mcp.config import load_config
from kb_mcp.events.store import EventStore
from kb_mcp.learning.models import ResolverInput
from kb_mcp.learning.packet_builder import build_learning_packet
from kb_mcp.learning.resolver import resolve_learning_assets
from kb_mcp.learning.revocation import expire_learning_assets, retract_learning_asset, supersede_learning_asset


class LearningRevocationTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmpdir.cleanup)
        root = Path(self.tmpdir.name)
        self.config_dir = root / "config"
        self.vault = root / "vault"
        os.environ["KB_CONFIG_DIR"] = str(self.config_dir)
        self.addCleanup(lambda: os.environ.pop("KB_CONFIG_DIR", None))
        self.vault.mkdir(parents=True, exist_ok=True)
        (self.vault / "projects" / "demo" / "gap").mkdir(parents=True, exist_ok=True)
        cfg_path = self.config_dir / "config.yml"
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(
            yaml.safe_dump(
                {"vault_path": str(self.vault), "kb_root": "", "timezone": "Asia/Tokyo", "obsidian_cli": "auto", "vault_git": False}
            ),
            encoding="utf-8",
        )
        load_config.cache_clear()

    def test_retract_learning_asset_invalidates_packets_and_hides_asset(self) -> None:
        store = EventStore()
        self._seed_asset(store, asset_key="asset-1")
        request = ResolverInput(source_tool="kb", source_client="kb-mcp", project="demo")
        packet = build_learning_packet(request, tool_name="gap", assets=resolve_learning_assets(request, store=store), store=store)

        result = retract_learning_asset(asset_key="asset-1", actor="tester", reason="bad learning", store=store)

        self.assertEqual(result["lifecycle"], "retracted")
        self.assertEqual(result["invalidated_packets"], 1)
        row = store.get_learning_asset("asset-1")
        self.assertIsNotNone(row)
        self.assertEqual(row["lifecycle"], "retracted")
        self.assertEqual(row["learning_state_visibility"], "retracted")
        packet_row = store.get_learning_packet(str(packet["packet_id"]))
        self.assertEqual(packet_row["status"], "invalidated")
        self.assertEqual(len(store.list_learning_revocations()), 1)
        self.assertEqual(resolve_learning_assets(request, store=store), [])

    def test_supersede_learning_asset_marks_old_asset_stale(self) -> None:
        store = EventStore()
        self._seed_asset(store, asset_key="asset-old", updated_at="2026-03-26T00:00:00+00:00")
        self._seed_asset(store, asset_key="asset-new", updated_at="2026-03-26T00:10:00+00:00")
        request = ResolverInput(source_tool="kb", source_client="kb-mcp", project="demo")
        build_learning_packet(request, tool_name="gap", assets=resolve_learning_assets(request, store=store), store=store)

        result = supersede_learning_asset(
            asset_key="asset-old",
            replacement_asset_key="asset-new",
            actor="tester",
            reason="newer rule wins",
            store=store,
        )

        self.assertEqual(result["lifecycle"], "superseded")
        old_row = store.get_learning_asset("asset-old")
        self.assertEqual(old_row["confidence"], "stale")
        resolved = resolve_learning_assets(request, store=store)
        self.assertEqual([item.asset_key for item in resolved], ["asset-new"])

    def test_expire_learning_assets_marks_old_rows_expired(self) -> None:
        store = EventStore()
        self._seed_asset(store, asset_key="asset-old", updated_at="2026-03-26T00:00:00+00:00")
        self._seed_asset(store, asset_key="asset-new", updated_at="2026-03-26T00:30:00+00:00")
        request = ResolverInput(source_tool="kb", source_client="kb-mcp", project="demo")
        build_learning_packet(request, tool_name="gap", assets=resolve_learning_assets(request, store=store), store=store)

        result = expire_learning_assets(
            before="2026-03-26T00:05:00+00:00",
            actor="tester",
            reason="ttl elapsed",
            store=store,
        )

        self.assertEqual(result["expired"], 1)
        old_row = store.get_learning_asset("asset-old")
        self.assertEqual(old_row["lifecycle"], "expired")
        resolved = resolve_learning_assets(request, store=store)
        self.assertEqual([item.asset_key for item in resolved], ["asset-new"])

    def _seed_asset(
        self,
        store: EventStore,
        *,
        asset_key: str,
        updated_at: str = "2026-03-26T00:00:00+00:00",
    ) -> None:
        store.upsert_learning_asset(
            asset_key=asset_key,
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope="project_local",
            force="preferred",
            confidence="reviewed",
            lifecycle="active",
            provenance={"project": "demo"},
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
            created_at="2026-03-26T00:00:00+00:00",
            updated_at=updated_at,
        )



from __future__ import annotations

import io
import json
import os
import tempfile
import unittest
from contextlib import redirect_stdout
from pathlib import Path
from unittest.mock import patch

import yaml

from kb_mcp.config import load_config
from kb_mcp.events.store import EventStore


class LearningCliTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmpdir.cleanup)
        root = Path(self.tmpdir.name)
        self.config_dir = root / "config"
        self.vault = root / "vault"
        os.environ["KB_CONFIG_DIR"] = str(self.config_dir)
        self.addCleanup(lambda: os.environ.pop("KB_CONFIG_DIR", None))
        self.vault.mkdir(parents=True, exist_ok=True)
        cfg_path = self.config_dir / "config.yml"
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(
            yaml.safe_dump(
                {"vault_path": str(self.vault), "kb_root": "", "timezone": "Asia/Tokyo", "obsidian_cli": "auto", "vault_git": False}
            ),
            encoding="utf-8",
        )
        load_config.cache_clear()

    def test_judge_learning_state_prints_assets(self) -> None:
        store = EventStore()
        store.upsert_learning_asset(
            asset_key="asset-1",
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope="project_local",
            force="preferred",
            confidence="reviewed",
            lifecycle="active",
            provenance={"project": "demo"},
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
        )
        from kb_mcp import cli

        buf = io.StringIO()
        with patch("sys.argv", ["kb-mcp", "judge", "learning-state", "--limit", "10"]):
            with redirect_stdout(buf):
                cli.main()
        payload = json.loads(buf.getvalue())
        self.assertEqual(payload["total"], 1)
        self.assertEqual(payload["results"][0]["asset_key"], "asset-1")

    def test_judge_retract_learning_updates_asset(self) -> None:
        store = EventStore()
        store.upsert_learning_asset(
            asset_key="asset-active",
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope="project_local",
            force="preferred",
            confidence="reviewed",
            lifecycle="active",
            provenance={"project": "demo"},
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
        )
        from kb_mcp import cli

        buf = io.StringIO()
        with patch("sys.argv", ["kb-mcp", "judge", "retract-learning", "asset-active", "--reason", "wrong"]):
            with redirect_stdout(buf):
                cli.main()
        payload = json.loads(buf.getvalue())
        self.assertEqual(payload["lifecycle"], "retracted")
        row = store.get_learning_asset("asset-active")
        self.assertEqual(row["lifecycle"], "retracted")



"""Resolve applicable learning assets for one runtime request."""

from __future__ import annotations

import json
from typing import Iterable

from kb_mcp.events.store import EventStore
from kb_mcp.learning.models import LearningAssetView, ResolverInput
from kb_mcp.resolver import resolve_project

_SCOPE_ORDER = {
    "session_local": 0,
    "client_local": 1,
    "project_local": 2,
    "user_global": 3,
    "general": 4,
}
_CONFIDENCE_ORDER = {
    "stable": 0,
    "reviewed": 1,
    "candidate": 2,
    "observed": 3,
    "stale": 4,
}
_FORCE_ORDER = {
    "guardrail": 0,
    "default": 1,
    "preferred": 2,
    "hint": 3,
}
_APPLICABLE_LIFECYCLES = {"active", "candidate"}


def resolve_learning_assets(
    request: ResolverInput,
    *,
    store: EventStore | None = None,
) -> list[LearningAssetView]:
    resolved_project, _ = resolve_project(
        project=request.project,
        cwd=request.cwd,
        repo=request.repo,
    )
    rows = (store or EventStore()).list_learning_assets()
    matched = [
        _row_to_view(row)
        for row in rows
        if _matches_scope(_row_to_view(row), request=request, resolved_project=resolved_project)
    ]
    return sorted(
        matched,
        key=lambda item: (
            _SCOPE_ORDER[item.scope],
            _CONFIDENCE_ORDER[item.confidence],
            _FORCE_ORDER[item.force],
            -_updated_rank(item.updated_at),
            item.asset_key,
        ),
    )


def _matches_scope(
    asset: LearningAssetView,
    *,
    request: ResolverInput,
    resolved_project: str | None,
) -> bool:
    if asset.lifecycle not in _APPLICABLE_LIFECYCLES:
        return False
    provenance = asset.provenance
    if asset.scope == "session_local":
        return bool(request.session_id) and provenance.get("session_id") == request.session_id
    if asset.scope == "client_local":
        return provenance.get("source_client") == request.source_client
    if asset.scope == "project_local":
        return bool(resolved_project) and provenance.get("project") == resolved_project
    if asset.scope in {"user_global", "general"}:
        return True
    return False


def _row_to_view(row: object) -> LearningAssetView:
    provenance = _json_obj(getattr(row, "__getitem__")("provenance_json"))
    traceability = _json_obj(getattr(row, "__getitem__")("traceability_json"))
    revocation_path = _json_obj(getattr(row, "__getitem__")("revocation_path_json"))
    return LearningAssetView(
        asset_key=str(getattr(row, "__getitem__")("asset_key")),
        memory_class=str(getattr(row, "__getitem__")("memory_class")),
        update_target=str(getattr(row, "__getitem__")("update_target")),
        scope=str(getattr(row, "__getitem__")("scope")),
        force=str(getattr(row, "__getitem__")("force")),
        confidence=str(getattr(row, "__getitem__")("confidence")),
        lifecycle=str(getattr(row, "__getitem__")("lifecycle")),
        learning_state_visibility=str(getattr(row, "__getitem__")("learning_state_visibility")),
        candidate_key=_maybe_str(getattr(row, "__getitem__")("candidate_key")),
        review_id=_maybe_str(getattr(row, "__getitem__")("review_id")),
        materialization_key=_maybe_str(getattr(row, "__getitem__")("materialization_key")),
        note_id=_maybe_str(getattr(row, "__getitem__")("note_id")),
        note_path=_maybe_str(getattr(row, "__getitem__")("note_path")),
        updated_at=str(getattr(row, "__getitem__")("updated_at")),
        provenance=provenance,
        traceability=traceability,
        revocation_path=revocation_path,
    )


def _json_obj(value: object) -> dict[str, object]:
    if not value:
        return {}
    loaded = json.loads(str(value))
    return loaded if isinstance(loaded, dict) else {}


def _maybe_str(value: object) -> str | None:
    return str(value) if value is not None else None


def _updated_rank(updated_at: str) -> int:
    digits = "".join(ch for ch in updated_at if ch.isdigit())
    return int(digits or "0")


def resolver_orders() -> dict[str, Iterable[str]]:
    return {
        "scope": _SCOPE_ORDER.keys(),
        "confidence": _CONFIDENCE_ORDER.keys(),
        "force": _FORCE_ORDER.keys(),
    }



from __future__ import annotations

import json
import os
import tempfile
import unittest
from pathlib import Path

import yaml

from kb_mcp.config import load_config
from kb_mcp.events.middleware import with_tool_events
from kb_mcp.events.store import EventStore


class LearningApplicationTraceTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmpdir.cleanup)
        root = Path(self.tmpdir.name)
        self.config_dir = root / "config"
        self.vault = root / "vault"
        os.environ["KB_CONFIG_DIR"] = str(self.config_dir)
        self.addCleanup(lambda: os.environ.pop("KB_CONFIG_DIR", None))
        self.vault.mkdir(parents=True, exist_ok=True)
        (self.vault / "projects" / "demo" / "gap").mkdir(parents=True, exist_ok=True)
        cfg_path = self.config_dir / "config.yml"
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(
            yaml.safe_dump(
                {"vault_path": str(self.vault), "kb_root": "", "timezone": "Asia/Tokyo", "obsidian_cli": "auto", "vault_git": False}
            ),
            encoding="utf-8",
        )
        load_config.cache_clear()

    def test_tool_middleware_records_packet_and_application(self) -> None:
        os.environ["KB_VENDOR_SESSION_ID"] = "session-trace"
        self.addCleanup(lambda: os.environ.pop("KB_VENDOR_SESSION_ID", None))
        store = EventStore()
        store.upsert_learning_asset(
            asset_key="asset-project",
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope="project_local",
            force="preferred",
            confidence="reviewed",
            lifecycle="active",
            provenance={"project": "demo"},
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
        )

        def sample(*, project: str) -> str:
            return project

        wrapped = with_tool_events("kb", "kb-mcp", "sample", sample)
        result = wrapped(project="demo", ctx=None)

        self.assertEqual(result, "demo")
        counts = store.learning_packet_counts()
        self.assertEqual(counts["packets"], 1)
        self.assertEqual(counts["applications"], 1)
        packet_id = None
        with store.transaction() as conn:
            packet_id = str(
                conn.execute(
                    "SELECT packet_id FROM learning_packets ORDER BY created_at DESC LIMIT 1"
                ).fetchone()["packet_id"]
            )
            row = conn.execute(
                "SELECT raw_payload_json FROM events WHERE aggregate_type='tool' AND event_name='tool_started' ORDER BY rowid DESC LIMIT 1"
            ).fetchone()
        packet_row = store.get_learning_packet(packet_id)
        payload = json.loads(row["raw_payload_json"])
        self.assertIn("packet_id", payload)
        self.assertIsNotNone(packet_row["expires_at"])



from __future__ import annotations

import os
import tempfile
import unittest
from pathlib import Path

import yaml

from kb_mcp.config import load_config
from kb_mcp.events.store import EventStore
from kb_mcp.learning.models import ResolverInput
from kb_mcp.learning.resolver import resolve_learning_assets


class LearningResolverTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmpdir.cleanup)
        root = Path(self.tmpdir.name)
        self.config_dir = root / "config"
        self.vault = root / "vault"
        self.project = "demo"
        os.environ["KB_CONFIG_DIR"] = str(self.config_dir)
        self.addCleanup(lambda: os.environ.pop("KB_CONFIG_DIR", None))
        self.vault.mkdir(parents=True, exist_ok=True)
        for subdir in [
            "projects/demo/session-log",
            "projects/demo/draft",
            "projects/demo/adr",
            "projects/demo/gap",
            "projects/demo/knowledge",
            "general/knowledge",
            "general/requirements",
            "inbox",
        ]:
            (self.vault / subdir).mkdir(parents=True, exist_ok=True)
        cfg_path = self.config_dir / "config.yml"
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(
            yaml.safe_dump(
                {
                    "vault_path": str(self.vault),
                    "kb_root": "",
                    "timezone": "Asia/Tokyo",
                    "obsidian_cli": "auto",
                    "vault_git": False,
                }
            ),
            encoding="utf-8",
        )
        load_config.cache_clear()

    def test_resolver_orders_scopes_narrow_first(self) -> None:
        store = EventStore()
        self._seed_asset(
            store,
            asset_key="asset-general",
            scope="general",
            confidence="stable",
            force="default",
            provenance={"project": self.project, "source_client": "codex-cli", "session_id": "session-1"},
        )
        self._seed_asset(
            store,
            asset_key="asset-project",
            scope="project_local",
            confidence="reviewed",
            force="preferred",
            provenance={"project": self.project, "source_client": "codex-cli", "session_id": "session-1"},
        )
        self._seed_asset(
            store,
            asset_key="asset-client",
            scope="client_local",
            confidence="reviewed",
            force="preferred",
            provenance={"project": self.project, "source_client": "codex-cli", "session_id": "session-1"},
        )
        self._seed_asset(
            store,
            asset_key="asset-session",
            scope="session_local",
            confidence="candidate",
            force="hint",
            provenance={"project": self.project, "source_client": "codex-cli", "session_id": "session-1"},
        )

        resolved = resolve_learning_assets(
            ResolverInput(
                source_tool="codex",
                source_client="codex-cli",
                session_id="session-1",
                project=self.project,
            ),
            store=store,
        )

        self.assertEqual(
            [item.asset_key for item in resolved],
            ["asset-session", "asset-client", "asset-project", "asset-general"],
        )

    def test_resolver_uses_confidence_then_force_then_updated_at(self) -> None:
        store = EventStore()
        self._seed_asset(
            store,
            asset_key="asset-stable",
            scope="project_local",
            confidence="stable",
            force="hint",
            updated_at="2026-03-26T00:00:00+00:00",
            provenance={"project": self.project},
        )
        self._seed_asset(
            store,
            asset_key="asset-guardrail",
            scope="project_local",
            confidence="reviewed",
            force="guardrail",
            updated_at="2026-03-26T00:10:00+00:00",
            provenance={"project": self.project},
        )
        self._seed_asset(
            store,
            asset_key="asset-newer",
            scope="project_local",
            confidence="reviewed",
            force="preferred",
            updated_at="2026-03-26T00:20:00+00:00",
            provenance={"project": self.project},
        )

        resolved = resolve_learning_assets(
            ResolverInput(source_tool="codex", source_client="codex-cli", project=self.project),
            store=store,
        )

        self.assertEqual(
            [item.asset_key for item in resolved],
            ["asset-stable", "asset-guardrail", "asset-newer"],
        )

    def test_resolver_filters_by_scope_context(self) -> None:
        store = EventStore()
        self._seed_asset(
            store,
            asset_key="asset-other-session",
            scope="session_local",
            provenance={"project": self.project, "source_client": "codex-cli", "session_id": "session-2"},
        )
        self._seed_asset(
            store,
            asset_key="asset-other-client",
            scope="client_local",
            provenance={"project": self.project, "source_client": "claude-code"},
        )
        self._seed_asset(
            store,
            asset_key="asset-other-project",
            scope="project_local",
            provenance={"project": "another-project"},
        )
        self._seed_asset(
            store,
            asset_key="asset-match",
            scope="project_local",
            provenance={"project": self.project},
        )

        resolved = resolve_learning_assets(
            ResolverInput(
                source_tool="codex",
                source_client="codex-cli",
                session_id="session-1",
                project=self.project,
            ),
            store=store,
        )

        self.assertEqual([item.asset_key for item in resolved], ["asset-match"])

    def _seed_asset(
        self,
        store: EventStore,
        *,
        asset_key: str,
        scope: str,
        confidence: str = "reviewed",
        force: str = "preferred",
        updated_at: str = "2026-03-26T00:00:00+00:00",
        provenance: dict[str, object],
    ) -> None:
        store.upsert_learning_asset(
            asset_key=asset_key,
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope=scope,
            force=force,
            confidence=confidence,
            lifecycle="active",
            provenance=provenance,
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
            created_at="2026-03-26T00:00:00+00:00",
            updated_at=updated_at,
        )



from __future__ import annotations

import os
import tempfile
import unittest
from pathlib import Path

import yaml

from kb_mcp.config import load_config
from kb_mcp.events.store import EventStore
from kb_mcp.learning.models import ResolverInput
from kb_mcp.learning.packet_builder import build_learning_packet
from kb_mcp.learning.resolver import resolve_learning_assets


class LearningPacketBuilderTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmpdir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmpdir.cleanup)
        root = Path(self.tmpdir.name)
        self.config_dir = root / "config"
        self.vault = root / "vault"
        os.environ["KB_CONFIG_DIR"] = str(self.config_dir)
        self.addCleanup(lambda: os.environ.pop("KB_CONFIG_DIR", None))
        self.vault.mkdir(parents=True, exist_ok=True)
        (self.vault / "projects" / "demo" / "gap").mkdir(parents=True, exist_ok=True)
        cfg_path = self.config_dir / "config.yml"
        cfg_path.parent.mkdir(parents=True, exist_ok=True)
        cfg_path.write_text(
            yaml.safe_dump(
                {"vault_path": str(self.vault), "kb_root": "", "timezone": "Asia/Tokyo", "obsidian_cli": "auto", "vault_git": False}
            ),
            encoding="utf-8",
        )
        load_config.cache_clear()

    def test_build_learning_packet_persists_packet_and_assets(self) -> None:
        store = EventStore()
        store.upsert_learning_asset(
            asset_key="asset-project",
            candidate_key=None,
            review_id=None,
            materialization_key=None,
            note_id=None,
            note_path=None,
            memory_class="gap",
            update_target="behavior_style",
            scope="project_local",
            force="preferred",
            confidence="reviewed",
            lifecycle="active",
            provenance={"project": "demo"},
            traceability={},
            revocation_path={},
            learning_state_visibility="active",
            source_status="materialized",
        )
        request = ResolverInput(source_tool="kb", source_client="kb-mcp", project="demo")
        assets = resolve_learning_assets(request, store=store)

        packet = build_learning_packet(request, tool_name="gap", assets=assets, store=store)

        self.assertIsNotNone(packet)
        counts = store.learning_packet_counts()
        self.assertEqual(counts["packets"], 1)
        packet_row = store.get_learning_packet(str(packet["packet_id"]))
        self.assertEqual(packet_row["status"], "active")
        self.assertIsNotNone(packet_row["expires_at"])



"""Learning asset revocation helpers."""

from __future__ import annotations

from datetime import UTC, datetime, timedelta
import json

from kb_mcp.events.store import EventStore
from kb_mcp.note import generate_ulid

DEFAULT_PACKET_TTL_SECONDS = 3600


def packet_expires_at(*, now: datetime | None = None, ttl_seconds: int = DEFAULT_PACKET_TTL_SECONDS) -> str:
    base = now or datetime.now(UTC)
    return (base + timedelta(seconds=ttl_seconds)).isoformat()


def retract_learning_asset(
    *,
    asset_key: str,
    actor: str,
    reason: str,
    store: EventStore | None = None,
) -> dict[str, object]:
    active_store = store or EventStore()
    row = active_store.get_learning_asset(asset_key)
    if row is None:
        raise ValueError(f"learning asset not found: {asset_key}")
    current = _json_obj(row["revocation_path_json"])
    invalidated = active_store.invalidate_learning_packets(
        asset_keys=[asset_key],
        reason=f"retract:{asset_key}",
    )
    updated = active_store.update_learning_asset(
        asset_key=asset_key,
        lifecycle="retracted",
        learning_state_visibility="retracted",
        revocation_path={
            **current,
            "action": "retract",
            "actor": actor,
            "reason": reason,
            "retracted_at": _now_iso(),
        },
        source_status="revoked",
        confidence="stale",
    )
    active_store.record_learning_revocation(
        revocation_id=generate_ulid(),
        asset_key=asset_key,
        action="retract",
        actor=actor,
        reason=reason,
        replacement_asset_key=None,
        invalidated_packet_count=invalidated,
    )
    return _result(updated, invalidated)


def supersede_learning_asset(
    *,
    asset_key: str,
    replacement_asset_key: str,
    actor: str,
    reason: str,
    store: EventStore | None = None,
) -> dict[str, object]:
    active_store = store or EventStore()
    row = active_store.get_learning_asset(asset_key)
    if row is None:
        raise ValueError(f"learning asset not found: {asset_key}")
    replacement = active_store.get_learning_asset(replacement_asset_key)
    if replacement is None:
        raise ValueError(f"replacement learning asset not found: {replacement_asset_key}")
    current = _json_obj(row["revocation_path_json"])
    invalidated = active_store.invalidate_learning_packets(
        asset_keys=[asset_key],
        reason=f"supersede:{asset_key}",
    )
    updated = active_store.update_learning_asset(
        asset_key=asset_key,
        lifecycle="superseded",
        learning_state_visibility="superseded",
        revocation_path={
            **current,
            "action": "supersede",
            "actor": actor,
            "reason": reason,
            "superseded_by": replacement_asset_key,
            "superseded_at": _now_iso(),
        },
        source_status="superseded",
        confidence="stale",
    )
    active_store.record_learning_revocation(
        revocation_id=generate_ulid(),
        asset_key=asset_key,
        action="supersede",
        actor=actor,
        reason=reason,
        replacement_asset_key=replacement_asset_key,
        invalidated_packet_count=invalidated,
    )
    return _result(updated, invalidated)


def expire_learning_assets(
    *,
    before: str,
    actor: str,
    reason: str,
    limit: int = 100,
    store: EventStore | None = None,
) -> dict[str, object]:
    active_store = store or EventStore()
    rows = active_store.list_expirable_learning_assets(before=before, limit=limit)
    results: list[dict[str, object]] = []
    for row in rows:
        asset_key = str(row["asset_key"])
        invalidated = active_store.invalidate_learning_packets(
            asset_keys=[asset_key],
            reason=f"expire:{asset_key}",
        )
        updated = active_store.update_learning_asset(
            asset_key=asset_key,
            lifecycle="expired",
            learning_state_visibility="expired",
            revocation_path={
                **_json_obj(row["revocation_path_json"]),
                "action": "expire",
                "actor": actor,
                "reason": reason,
                "expired_at": _now_iso(),
            },
            source_status="expired",
            confidence="stale",
        )
        active_store.record_learning_revocation(
            revocation_id=generate_ulid(),
            asset_key=asset_key,
            action="expire",
            actor=actor,
            reason=reason,
            replacement_asset_key=None,
            invalidated_packet_count=invalidated,
        )
        results.append(_result(updated, invalidated))
    return {
        "expired": len(results),
        "results": results,
    }


def invalidate_expired_packets(
    *,
    store: EventStore | None = None,
) -> dict[str, int]:
    count = (store or EventStore()).invalidate_expired_learning_packets()
    return {"invalidated_packets": count}


def _result(row: object, invalidated_packets: int) -> dict[str, object]:
    return {
        "asset_key": str(getattr(row, "__getitem__")("asset_key")),
        "lifecycle": str(getattr(row, "__getitem__")("lifecycle")),
        "learning_state_visibility": str(getattr(row, "__getitem__")("learning_state_visibility")),
        "invalidated_packets": invalidated_packets,
    }


def _json_obj(value: object) -> dict[str, object]:
    if not value:
        return {}
    loaded = json.loads(str(value))
    return loaded if isinstance(loaded, dict) else {}


def _now_iso() -> str:
    return datetime.now(UTC).isoformat()
